use std::process::Stdio;

use bytes::Bytes;
use futures::FutureExt;
use serde::Deserialize;
use tokio::process::Command;
use tokio_stream::StreamExt;
use tokio_util::codec::{BytesCodec, FramedRead};

use super::Ctx;

#[derive(Debug, Deserialize)]
pub struct ExecReq {
    user: String,
    program: String,
    #[serde(default)]
    args: Vec<String>,
}

impl TryFrom<Bytes> for ExecReq {
    type Error = serde_json::Error;

    fn try_from(value: Bytes) -> Result<Self, Self::Error> {
        serde_json::from_slice(&value[..])
    }
}

/// An operating system program execution.
pub async fn exec(ctx: Ctx<ExecReq>) -> anyhow::Result<()> {
    // TODO: Tasks should be idempontent
    // TODO: Root user should be able to run only specific programs

    let mut cmd = if &ctx.body.user != "root" {
        let mut cmd = Command::new("sudo");
        cmd.arg("-iu")
            .arg(&ctx.body.user)
            .arg("--")
            .arg(&ctx.body.program);
        cmd
    } else {
        Command::new(&ctx.body.program)
    };

    let mut io = cmd
        .args(&ctx.body.args)
        .stdout(Stdio::piped())
        .stderr(Stdio::piped())
        .spawn()?;

    let stdout =
        FramedRead::new(io.stdout.take().unwrap(), BytesCodec::new()).filter_map(|stdout| {
            stdout
                .map(|bytes| bytes.freeze())
                .map_err(|err| tracing::error!(%err, "read error on stdout"))
                .ok()
        });

    let stderr =
        FramedRead::new(io.stderr.take().unwrap(), BytesCodec::new()).filter_map(|stderr| {
            stderr
                .map(|bytes| bytes.freeze())
                .map_err(|err| tracing::error!(%err, "read error on stderr"))
                .ok()
        });

    let exit = async move {
        io.wait()
            .await
            .map(|exit| {
                let exit = exit.to_string();
                Bytes::from(exit)
            })
            .unwrap()
    }
    .into_stream();

    //     Ok(Box::pin(stdout.merge(stderr).chain(exit)))
    Ok(())
}
